

#include <SDL2/SDL.h>

#include <chrono>
#include <format>
#include <memory>
#include <print>
#include <random>
#include <source_location>
#include <stdexcept>
#include <thread>

#ifdef __EMSCRIPTEN__
	#include <emscripten.h>
	#include <emscripten/html5.h>
#endif

namespace {

auto sdl_throw_current_error(const std::source_location location) {
	throw std::runtime_error{std::format(
		"error at {}:{}:{} {}", location.file_name(), location.line(),
		location.column(), SDL_GetError()
	)};
}

auto sdl_check(
	int errc,
	const std::source_location location = std::source_location::current()
) {
	if (errc != 0) { sdl_throw_current_error(location); }
}

// make sure someone already owns passed sdl_obj, or memory will be leaked
auto sdl_check(
	auto* sdl_obj,
	const std::source_location location = std::source_location::current()
) {
	if (sdl_obj == nullptr) { sdl_throw_current_error(location); }
	return sdl_obj;
}

template <typename SDLT, auto deletef>
struct sdl_wrapper_t : std::unique_ptr<SDLT, decltype(deletef)> {
	explicit sdl_wrapper_t(SDLT* p)
		: std::unique_ptr<SDLT, decltype(deletef)>{p, deletef} {
		sdl_check(p);
	}

	operator SDLT*() { return std::unique_ptr<SDLT, decltype(deletef)>::get(); }
};

using window_t = sdl_wrapper_t<SDL_Window, SDL_DestroyWindow>;
using surface_t = sdl_wrapper_t<SDL_Surface, SDL_FreeSurface>;
using renderer_t = sdl_wrapper_t<SDL_Renderer, SDL_DestroyRenderer>;

auto operator+(SDL_Point a, SDL_Point b) {
	return SDL_Point{.x = a.x + b.x, .y = a.y + b.y};
}

constexpr int width = 1000;
constexpr int height = 500;

// constexpr int width = 3440;
// constexpr int height = 1440;

template <auto EF>
struct [[nodiscard("give this a name so SDL_Quit is called at the end"
)]] Defer {
	Defer() = default;
	Defer(const Defer&) = delete;
	Defer(Defer&&) = delete;
	auto operator=(Defer&&) = delete;
	auto operator=(const Defer&) = delete;

	~Defer() { EF(); }
};

using clock = std::chrono::steady_clock;

auto poll_events(bool& continu) {
	for (SDL_Event e; SDL_PollEvent(&e);) {
		if (e.type == SDL_QUIT) { continu = false; }
	}
}
} // namespace

template <typename Loopdata>
void mainloop(void* userData) {
	constexpr SDL_Point directions[] = {
		{.x = -1, .y = 0}, {.x = 1, .y = 0}, {.x = 0, .y = -1}, {.x = 0, .y = 1}
	};
	constexpr auto batchsize = 1000;
	auto& [renderer, texture, pos, pos1, rne, dist, start_time, next_poll_events_time, next_frame_render_time, frame_start, frame_time, continu] =
		*static_cast<Loopdata*>(userData);

	// measure frame_time
	const auto now = clock::now();
	frame_time = now - frame_start;
	frame_start = now;

	if (now >= next_poll_events_time) {
		next_poll_events_time = now + std::chrono::milliseconds{200};
		poll_events(continu);
		std::println("last frame_time={} batchsize={}", frame_time, batchsize);
	}

	std::array<SDL_Point, batchsize> point_batch;
	for (auto step = 0z; step < batchsize; ++step) {
		SDL_Point newpoint;
		do {
			newpoint = pos + directions[dist(rne)];
		} while (newpoint.x < 0 or newpoint.x >= width or newpoint.y < 0 or
		         newpoint.y >= height);
		pos = newpoint;
		point_batch[step] = newpoint;
	}

	///// Render image
	sdl_check(SDL_SetRenderTarget(renderer, texture));
	sdl_check(SDL_SetRenderDrawBlendMode(
		renderer,
		SDL_ComposeCustomBlendMode(
			SDL_BLENDFACTOR_SRC_ALPHA, SDL_BLENDFACTOR_ONE_MINUS_SRC_ALPHA,
			SDL_BLENDOPERATION_ADD, SDL_BLENDFACTOR_ZERO, SDL_BLENDFACTOR_ONE,
			SDL_BLENDOPERATION_ADD
		)
	));
	sdl_check(SDL_SetRenderDrawColor(renderer, 0x00, 0x00, 0x00, 32));
	sdl_check(
		SDL_RenderDrawPoints(renderer, point_batch.data(), point_batch.size())
	);
	for (auto step = 0z; step < batchsize; ++step) {
		SDL_Point newpoint;
		do {
			newpoint = pos1 + directions[dist(rne)];
		} while (newpoint.x < 0 or newpoint.x >= width or newpoint.y < 0 or
		         newpoint.y >= height);
		pos1 = newpoint;
		point_batch[step] = newpoint;
	}
	sdl_check(SDL_SetRenderDrawColor(renderer, 0xff, 0xff, 0xff, 32));
	sdl_check(
		SDL_RenderDrawPoints(renderer, point_batch.data(), point_batch.size())
	);
	// sdl_check(SDL_SetRenderDrawColor(renderer, 0xff, 0xff, 0xff,
	// 32)); sdl_check(SDL_RenderFillRect(renderer, nullptr));
	// SDL_RenderCopy(renderer, fade_texture, nullptr, nullptr);

	///// Render screen
	sdl_check(SDL_SetRenderTarget(renderer, nullptr));
	sdl_check(
		SDL_SetRenderDrawColor(renderer, 0x00, 0x00, 0x00, SDL_ALPHA_OPAQUE)
	);
	SDL_RenderClear(renderer);
	sdl_check(SDL_SetRenderDrawBlendMode(renderer, SDL_BLENDMODE_NONE));
	// sdl_check(SDL_SetTextureBlendMode(texture, SDL_BLENDMODE_NONE));
	sdl_check(SDL_RenderCopy(renderer, texture, nullptr, nullptr));
	SDL_RenderPresent(renderer);
}

int main() try {
	sdl_check(SDL_Init(SDL_INIT_VIDEO));
	Defer<SDL_Quit> defer_SDL_Quit;

	window_t window{SDL_CreateWindow(
		"random walk", SDL_WINDOWPOS_UNDEFINED, SDL_WINDOWPOS_UNDEFINED, width,
		height, 0 // SDL_WINDOW_FULLSCREEN
	)};
	renderer_t renderer{SDL_CreateRenderer(window, -1, SDL_RENDERER_ACCELERATED)
	};
	SDL_SetHint(SDL_HINT_RENDER_SCALE_QUALITY, "1");
	{
		// dump renderer info
		SDL_RendererInfo renderer_info;
		SDL_GetRendererInfo(renderer, &renderer_info);
		std::println("Renderer name: {}", renderer_info.name);
		std::println("Texture formats: ");
		for (Uint32 i = 0; i < renderer_info.num_texture_formats; i++) {
			std::puts(SDL_GetPixelFormatName(renderer_info.texture_formats[i]));
		}
	}

	// no clue who owns this
	auto* texture = SDL_CreateTexture(
		renderer, SDL_PIXELFORMAT_RGB888, SDL_TEXTUREACCESS_TARGET, width,
		height
	);
	sdl_check(SDL_SetRenderTarget(renderer, texture));
	sdl_check(
		SDL_SetRenderDrawColor(renderer, 0xff, 0xff, 0xff, SDL_ALPHA_OPAQUE)
	);
	sdl_check(SDL_RenderClear(renderer));

	SDL_Point pos{.x = width / 2, .y = height / 2};
	SDL_Point pos1{.x = width / 2, .y = height / 2};

	std::mt19937_64 rne(std::random_device{}());
	std::uniform_int_distribution dist(0, 3);
	bool continu = true;

	const auto start_time = clock::now();
	auto next_poll_events_time = start_time;
	auto next_frame_render_time = start_time;
	auto frame_start = start_time;
	clock::duration frame_time;
	auto loopdata = std::tie(
		renderer, texture, pos, pos1, rne, dist, start_time,
		next_poll_events_time, next_frame_render_time, frame_start, frame_time,
		continu
	);
	std::puts("initialization complete");
#ifdef __EMSCRIPTEN__
	// Receives a function to call and some user data to provide it.
	emscripten_set_main_loop_arg(mainloop<decltype(loopdata)>, &loopdata, 0, 1);
#else
	while (continu) {
		mainloop<decltype(loopdata)>(&loopdata);
		std::this_thread::sleep_until(next_frame_render_time);
		next_frame_render_time += std::chrono::nanoseconds{16666666};
	}
#endif

	// leaking memory, nothing I can do
} catch (const std::exception& e) {

	std::puts(e.what());
	return EXIT_FAILURE;
}